#
# Copyright 2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# This was migrated from muSpectre byy Antoine Sanner. Lars Pastewka is author


import numbers
import numpy as np

from netCDF4 import Dataset

import NuMPI


class NCStructuredGridFrame(object):
    def __init__(self, parent, index):
        self._parent = parent
        self._index = index
        if self._index < 0:
            self._index += len(parent)

    def _create_variable(self, name, value):
        if isinstance(value, numbers.Integral):
            self.parent._data.createVariable(name, 'i4', ('frame',))
        elif isinstance(value, numbers.Real):
            self.parent._data.createVariable(name, 'f8', ('frame',))
        else:
            self.parent._create_variable(name, value, prefix=['frame'])

    def _variable(self, name, template):
        if name not in self.parent._data.variables:
            self._create_variable(name, template)
        var = self.parent._data.variables[name]
        if self.parent.is_parallel:
            pass
            # Collective I/O causes trouble on some configurations. Disable
            # for now.
            # var.set_collective(True)
        return var

    def __getattr__(self, name):
        if name[0] == '_':
            # FIXME: Use getattr here?
            return self.__dict__[name]

        var = self.parent._data.variables[name]
        if var.dimensions[0] != 'frame':
            raise ValueError("Variable '{}' exists in NetCDF file, but it "
                             "does not store per-frame data.".format(name))
        if self.decomposition == 'subdomain':
            if var.ndim == 1:  # it is a scalar
                return var[self._index]
            else:
                return var[self._index][self.subdomain_slices]
        else:
            return var[self._index]

    def __setattr__(self, name, value):
        if name[0] == '_':
            self.__dict__[name] = value
        elif isinstance(value, np.ndarray) and value.shape != ():
            var = self._variable(name, value)
            if self.decomposition == 'subdomain':
                var[tuple([self._index] + list(self.subdomain_slices))] = value
            else:
                var[self._index] = value
        else:
            var = self._variable(name, value)
            var[self._index] = value

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __setitem__(self, name, value):
        return self.__setattr__(name, value)

    @property
    def index(self):
        return self._index

    @property
    def parent(self):
        return self._parent

    @property
    def decomposition(self):
        return self.parent.decomposition

    @property
    def nb_domain_grid_pts(self):
        return self.parent.nb_domain_grid_pts

    @property
    def subdomain_locations(self):
        return self.parent.subdomain_locations

    @property
    def nb_subdomain_grid_pts(self):
        return self.parent.nb_subdomain_grid_pts

    @property
    def subdomain_slices(self):
        return self.parent.subdomain_slices

    def sync(self):
        self.parent.sync()


class NCStructuredGrid(object):
    """
    Parallel I/O of structured grid data into a NetCDF file. The object once
    instantiated automatically creates NetCDF variables from data assigned to
    attributes. Data can be written to the file globally or to a frame.

    Serial example
    --------------

        import numpy as np
        from muFFT.NetCDF import NCStructuredGrid
        nb_grid_pts = [12, 11]
        nc = NCStructuredGrid('example.nc', 'w', nb_grid_pts)
        grid_data = np.random.random(nb_grid_pts)
        # Create a new variable `random_data` in the NetCDF file and write
        # `grid_data` to it.
        nc.random_data = grid_data
        # Create a new variable `per_frame_data` with first unlimited
        # dimension 'frame'. Assign `grid_data` to frame 1. (Frame 0
        # will be undefined in this case.)
        nc[1].per_frame_data = grid_data

    This example create a file with the following structure:
        netcdf example {
        dimensions:
            frame = UNLIMITED ; // (2 currently)
            grid_x = 12 ;
            grid_y = 11 ;
        variables:
            double random_data(grid_x, grid_y) ;
            double per_frame_data(frame, grid_x, grid_y) ;
        }

    Parallel example
    ----------------

        import numpy as np
        from mpi4py import MPI
        from muFFT import FFT
        from muFFT.NetCDF import NCStructuredGrid
        comm = MPI.COMM_WORLD
        # Initialize muFFT object - we need this to get domain decomposition
        # information.
        nb_grid_pts = [25, 25, 25]
        fft = FFT(nb_grid_pts, communicator=comm)
        # Compute something, here we simply use random numbers.
        local_grid = np.random.random(fft.nb_subdomain_grid_pts)
        # Initialize the I/O object
        nc = NCStructuredGrid('example.nc', 'w',
                              nb_domain_grid_pts=fft.nb_domain_grid_pts,
                              decomposition='subdomain',
                              subdomain_locations=fft.subdomain_locations,
                              nb_subdomain_grid_pts=fft.nb_subdomain_grid_pts,
                              communicator=comm)
        # We only write the local portion of the grid data because the
        # decomposition is 'subdomain'.
        nc[0].grid_data = local_grid

    This example create a file with the following structure:
        netcdf example {
        dimensions:
            frame = UNLIMITED ; // (1 currently)
            grid_x = 25 ;
            grid_y = 25 ;
            grid_z = 25 ;
        variables:
            double grid_data(frame, grid_x, grid_y, grid_z) ;
        }
    """

    _program = 'NuMPI'
    _programVersion = NuMPI.__version__

    def __init__(self, fn, mode='r', nb_domain_grid_pts=None,
                 decomposition='serial', subdomain_locations=None,
                 nb_subdomain_grid_pts=None, communicator=None,
                 frame=0, format='NETCDF3_64BIT_DATA'):
        """
        Open a NetCDF file for reading or writing.

        Parameters
        ----------
        fn : str
            Name of the NetCDF file.
        mode : str
            Opening mode: 'r' - read, 'w' - write, 'a' - append.
        decomposition : str
            Specification of the data decomposition of the heights array. If
            set to 'subdomain', the grids contain only the part of the full
            grid local to the present MPI process. If set to 'domain', the
            grids contains the global array. Default: 'serial', which fails
            for parallel runs.
        nb_domain_grid_pts : tuple of ints
            Number of grid points for the full topography. This is only
            required if decomposition is set to 'subdomain'.
        subdomain_locations : tuple of ints
            Origin (location) of the subdomain handled by the present MPI
            process.
        nb_subdomain_grid_pts : tuple of ints
            Number of grid points within the subdomain handled by the present
            MPI process. This is only required if decomposition is set to
            'domain'.
        communicator : mpi4py communicator
            The MPI communicator object.
        frame : int
            Initial frame. (Default: 0)
        format : str
            NetCDF format string. (Default: 'NETCDF3_64BIT_DATA')
        """
        self._data = None

        self._fn = fn
        self._decomposition = decomposition
        self._communicator = communicator

        self._nb_domain_grid_pts = nb_domain_grid_pts
        self._subdomain_locations = subdomain_locations
        self._nb_subdomain_grid_pts = nb_subdomain_grid_pts

        # This is only a parallel run if more than one MPI process is spawned.
        # This test is necessary because on some configurations (unclear to me
        # exactly when) a call to `set_collective` complains with the error
        # "Parallel operation on file opened for non-parallel access" even
        # if there is only a single process even though the Dataset object was
        # instantiated with `parallel=True`.
        self._parallel = (self._communicator is not None) and (self._communicator.size > 1)

        if decomposition == 'serial':
            if self.is_parallel:
                raise RuntimeError("'serial' decomposition requested by the "
                                   "size of the communicator is {} and hence "
                                   "larger than 1."
                                   .format(self._communicator.size))
            self._subdomain_locations = (0, 0)
            self._nb_subdomain_grid_pts = nb_domain_grid_pts
        else:
            if subdomain_locations is None or nb_subdomain_grid_pts is None:
                raise ValueError('Please specify `subdomain_locations` and '
                                 '`nb_subdomain_grid_pts` if decomposition is '
                                 'not serial.')
            if communicator is None:
                raise ValueError("Please specify a communicator since you are "
                                 "requesting parallel I/O with decomposition "
                                 "'{}'.".format(decomposition))

        self._data = Dataset(fn, mode, format=format,
                             parallel=self.is_parallel,
                             comm=self._communicator)

        if mode[0] == 'w':
            if nb_domain_grid_pts is None:
                raise ValueError('Please specify the number of grid points '
                                 '`nb_domain_grid_pts` when creating a new '
                                 'file.')
            self._data.program = self._program
            self._data.programVersion = self._programVersion

            self._create_grid_dimensions(nb_domain_grid_pts)
        else:
            self._read_grid_dimensions()

        if frame < 0:
            self._cur_frame = len(self) + frame
        else:
            self._cur_frame = frame

    def __del__(self):
        if self._data is not None:
            self._data.close()
            self._data = None

    @staticmethod
    def _grid_dimension_name(i):
        return 'grid_' + chr(ord('x') + i)

    def _create_grid_dimensions(self, nb_domain_grid_pts):
        self._nb_domain_grid_pts = tuple(nb_domain_grid_pts)

        if 'frame' not in self._data.dimensions:
            self._data.createDimension('frame', None)
        for i, n in enumerate(nb_domain_grid_pts):
            dim_name = self._grid_dimension_name(i)
            if dim_name not in self._data.dimensions:
                self._data.createDimension(dim_name, n)

        self._data.sync()

    def _read_grid_dimensions(self):
        i = 0
        nb_domain_grid_pts = []

        dim_exists = True
        while dim_exists:
            try:
                dim = self._data.dimensions[self._grid_dimension_name(i)]
                nb_domain_grid_pts += [len(dim)]
                i += 1
            except KeyError:
                dim_exists = False

        self._nb_domain_grid_pts = tuple(nb_domain_grid_pts)

    def _tensor_dimension_name(self, n):
        """
        Return a variable for a tensor component of length `n`.
        """
        dim_name = 'tensor_{}'.format(n)
        if dim_name not in self._data.dimensions:
            self._data.createDimension(dim_name, n)
        return dim_name

    def _create_variable(self, name, template, prefix=[]):
        """
        Heuristics to guess dimensions for a new variable that can contain
        multidimension arrays of types given by `template`.
        """
        nb_grid_dims = len(self._nb_domain_grid_pts)
        guessed_nb_grid_pts = template.shape[:nb_grid_dims]
        if self._decomposition == 'subdomain' and \
                len(template.shape) >= nb_grid_dims and \
                guessed_nb_grid_pts == tuple(self._nb_subdomain_grid_pts):
            component_dims = [self._tensor_dimension_name(i)
                              for i in template.shape[nb_grid_dims:]]
            grid_dims = [self._grid_dimension_name(i)
                         for i in range(nb_grid_dims)]
            dims = grid_dims + component_dims
        elif len(template.shape) >= nb_grid_dims and \
                guessed_nb_grid_pts == tuple(self._nb_domain_grid_pts):
            component_dims = [self._tensor_dimension_name(i)
                              for i in template.shape[nb_grid_dims:]]
            grid_dims = [self._grid_dimension_name(i)
                         for i in range(nb_grid_dims)]
            dims = grid_dims + component_dims
        else:
            dims = [self._tensor_dimension_name(i) for i in template.shape]
        dims = prefix + dims
        self._data.createVariable(name, template.dtype.str, tuple(dims))

    def _variable(self, name, template, prefix=[]):
        if name not in self._data.variables:
            self._create_variable(name, template, prefix)
        var = self._data.variables[name]
        if self.is_parallel:
            pass
            # Collective I/O causes trouble on some configurations. Disable
            # for now.
            # var.set_collective(True)
        return var

    def __len__(self):
        return len(self._data.dimensions['frame'])

    def close(self):
        if self._data is not None:
            self._data.close()
            self._data = None

    def get_filename(self):
        return self._fn

    def get_next_frame(self):
        frame = NCStructuredGridFrame(self, self._cur_frame)
        self._cur_frame += 1
        return frame

    def set_cursor(self, cur_frame):
        self._cur_frame = cur_frame

    def get_cursor(self):
        return self._cur_frame

    def __getattr__(self, name):
        if name[0] == '_':
            return self.__dict__[name]

        if name in self._data.variables:
            var = self._data.variables[name]
            if self.decomposition == 'subdomain':
                return var[self.subdomain_slices]
            else:
                return var

        return getattr(self._data, name)

    def __setattr__(self, name, value):
        if name[0] == '_':
            self.__dict__[name] = value
        elif isinstance(value, np.ndarray) and value.shape != ():
            var = self._variable(name, value)
            if self.decomposition == 'subdomain':
                var[self.subdomain_slices] = value
            else:
                var[...] = value
        else:
            return setattr(self._data, name, value)

    def __setitem__(self, i, value):
        if isinstance(i, str):
            return self.__setattr__(i, value)
        raise RuntimeError('Cannot set full frame.')

    def __getitem__(self, i):
        if isinstance(i, str):
            return self.__getattr__(i)
        if isinstance(i, slice):
            return [NCStructuredGridFrame(self, j)
                    for j in range(*i.indices(len(self)))]
        return NCStructuredGridFrame(self, i)

    def __iter__(self):
        for i in range(len(self)):
            yield NCStructuredGridFrame(self, i)

    @property
    def decomposition(self):
        return self._decomposition

    @property
    def nb_domain_grid_pts(self):
        return self._nb_domain_grid_pts

    @property
    def subdomain_locations(self):
        return self._subdomain_locations

    @property
    def nb_subdomain_grid_pts(self):
        return self._nb_subdomain_grid_pts

    @property
    def subdomain_slices(self):
        return tuple(slice(start, start + length)
                     for start, length in zip(self.subdomain_locations,
                                              self.nb_subdomain_grid_pts))

    @property
    def is_parallel(self):
        return self._parallel

    def sync(self):
        self._data.sync()

    def __str__(self):
        return self._fn


###

def open(fn, mode='r', frame=None, **kwargs):
    if isinstance(fn, NCStructuredGrid):
        return fn
    i = fn.find('@')
    if i > 0:
        n = int(fn[i + 1:])
        fn = fn[:i]
        return NCStructuredGrid(fn, mode=mode, **kwargs)[n]
    elif frame is not None:
        return NCStructuredGrid(fn, mode=mode, **kwargs)[frame]
    else:
        return NCStructuredGrid(fn, mode=mode, **kwargs)



#
# Copyright 2018, 2020 Antoine Sanner
#           2020 k.o.haase@googlemail.com
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np
import os
import warnings

from NuMPI.IO.MPIFileIO import (
    save_npy, load_npy, MPIFileViewNPY, MPIFileTypeError, make_mpi_file_view
    )
from NuMPI import MPI
import NuMPI
import pytest

testdir = os.path.dirname(os.path.realpath(__file__))


def test_filesave_1D(comm):
    nb_domain_grid_pts = 128
    np.random.seed(2)
    globaldata = np.random.random(nb_domain_grid_pts)

    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    step = nb_domain_grid_pts // nprocs

    if rank == nprocs - 1:
        subdomain_slices = slice(rank * step, None)
        subdomain_locations = rank * step
        # nb_subdomain_grid_pts = nb_domain_grid_pts - rank * step
    else:
        subdomain_slices = slice(rank * step, (rank + 1) * step)
        subdomain_locations = rank * step
        # nb_subdomain_grid_pts = step

    localdata = globaldata[subdomain_slices]

    save_npy("test_Filesave_1D.npy",
             localdata,
             subdomain_locations,
             nb_domain_grid_pts,
             comm)
    comm.barrier()  # The MPI_File reading and closing doesn't have to
    # finish together
    loaded_data = np.load("test_Filesave_1D.npy")
    np.testing.assert_array_equal(loaded_data, globaldata)

    comm.barrier()
    if rank == 0:
        os.remove("test_Filesave_1D.npy")


@pytest.mark.skipif(MPI.COMM_WORLD.Get_size() > 1,
                    reason="test is only serial")
def test_filesave_1D_list():
    nb_domain_grid_pts = 8
    np.random.seed(2)
    globaldata = np.random.random(nb_domain_grid_pts).tolist()

    save_npy("test_Filesave_1D_list.npy", globaldata)

    loaded_data = np.load("test_Filesave_1D_list.npy")
    np.testing.assert_array_equal(loaded_data, globaldata)

    os.remove("test_Filesave_1D_list.npy")


@pytest.fixture(scope="module", params=("C", "F"))
def globaldata(request, comm):
    order = request.param

    rank = comm.Get_rank()

    nb_domain_grid_pts = (128, 128)
    np.random.seed(2)
    if order == "C":
        globaldata = np.random.random(nb_domain_grid_pts)
        assert globaldata.flags["C_CONTIGUOUS"]
    elif order == "F":
        globaldata = np.random.random(nb_domain_grid_pts[::-1]).transpose()
        assert globaldata.flags["F_CONTIGUOUS"]
    if rank == 0:
        np.save("test_FileLoad_2D.npy", globaldata)

    comm.barrier()

    yield globaldata
    comm.barrier()
    if rank == 0:
        os.remove("test_FileLoad_2D.npy")


# ## Helper function:

class DistributedData:
    def __init__(self, data, nb_domain_grid_pts, subdomain_locations):
        self.data = data
        self.nb_domain_grid_pts = nb_domain_grid_pts
        self.nb_subdomain_grid_pts = data.shape
        self.subdomain_locations = subdomain_locations

    @property
    def subdomain_slices(self):
        return tuple([slice(s, s + n) for s, n in
                      zip(self.subdomain_locations,
                          self.nb_subdomain_grid_pts)])


def make_2d_slab_y(comm, globaldata):
    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    nb_domain_grid_pts = globaldata.shape

    step = nb_domain_grid_pts[1] // nprocs
    if rank == nprocs - 1:
        subdomain_slices = (slice(None, None), slice(rank * step, None))
        subdomain_locations = [0, rank * step]
        # nb_subdomain_grid_pts = [nb_domain_grid_pts[0],
        #                          nb_domain_grid_pts[1] - rank * step]
    else:
        subdomain_slices = (
            slice(None, None), slice(rank * step, (rank + 1) * step))
        subdomain_locations = [0, rank * step]
        # nb_subdomain_grid_pts = [nb_domain_grid_pts[1], step]

    return DistributedData(globaldata[subdomain_slices],
                           nb_domain_grid_pts,
                           subdomain_locations)


def make_2d_slab_x(comm, globaldata):
    """
    returns the part of globaldata attribute to the present rank in 2D data
    decomposition
    Parameters
    ----------
    comm : communicator
    globaldata

    Returns
    -------
    DistributedData
    """
    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    nb_domain_grid_pts = globaldata.shape

    step = nb_domain_grid_pts[0] // nprocs

    if rank == nprocs - 1:
        subdomain_slices = (slice(rank * step, None), slice(None, None))
        subdomain_locations = [rank * step, 0]
        # nb_subdomain_grid_pts = [nb_domain_grid_pts[0] - rank * step,
        #                          nb_domain_grid_pts[1]]
    else:
        subdomain_slices = (
            slice(rank * step, (rank + 1) * step), slice(None, None))
        subdomain_locations = [rank * step, 0]
        # nb_subdomain_grid_pts = [step, nb_domain_grid_pts[1]]

    return DistributedData(globaldata[subdomain_slices], nb_domain_grid_pts,
                           subdomain_locations)


@pytest.mark.parametrize("decompfun", [make_2d_slab_x, make_2d_slab_y])
def test_FileSave_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    save_npy("test_Filesave_2D.npy",
             distdata.data,
             distdata.subdomain_locations,
             distdata.nb_domain_grid_pts, comm)
    comm.barrier()
    if comm.Get_rank() == 0:
        loaded_data = np.load("test_Filesave_2D.npy")
        np.testing.assert_array_equal(loaded_data, globaldata)

        os.remove("test_Filesave_2D.npy")
    comm.barrier()


@pytest.mark.parametrize("decompfun", [make_2d_slab_x, make_2d_slab_y])
def test_FileView_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    # arr = np.load("test_FileLoad_2D.npy")
    # assert arr.shape == self.nb_domain_grid_pts

    file = MPIFileViewNPY("test_FileLoad_2D.npy", comm=comm)

    assert file.nb_grid_pts == distdata.nb_domain_grid_pts
    assert file.dtype == globaldata.dtype

    loaded_data = file.read(
        nb_subdomain_grid_pts=distdata.nb_subdomain_grid_pts,
        subdomain_locations=distdata.subdomain_locations)

    np.testing.assert_array_equal(loaded_data, distdata.data)


@pytest.mark.parametrize("decompfun", [make_2d_slab_x, make_2d_slab_y])
def test_FileLoad_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    loaded_data = load_npy(
        "test_FileLoad_2D.npy",
        nb_subdomain_grid_pts=distdata.nb_subdomain_grid_pts,
        subdomain_locations=distdata.subdomain_locations,
        comm=comm)

    np.testing.assert_array_equal(loaded_data, distdata.data)


@pytest.fixture
def npyfile():
    """
    defines a filename and makes a cleanup once the test was executed
    Yields
    ------
    filename

    """
    yield "test_same_numpy.npy"
    try:
        os.remove("test_same_numpy.npy")
    except FileNotFoundError:
        pass


@pytest.mark.skip(reason="just some statements on numpy behaviour")
def test_detect_fortran_order(comm_self):
    # fix some statements on numpy behaviour

    # Prepare fortran array
    arr = np.array(range(6), dtype=float).reshape(3, 2)
    arr = arr.transpose()

    # States numpy behaviour
    assert arr.shape == (2, 3)
    assert arr.flags["C_CONTIGUOUS"] is False
    assert arr.flags["F_CONTIGUOUS"] is True

    np.save("test.npy", arr)

    # asserts the loaded array is exactly the same
    loaded = np.load("test.npy")
    assert loaded.shape == (2, 3)
    assert loaded.flags["C_CONTIGUOUS"] is False
    assert loaded.flags["F_CONTIGUOUS"] is True


def test_load_same_numpy_load(comm_self, npyfile):
    data = np.random.random(size=(2, 3))
    np.save(npyfile, data)
    loaded_data = load_npy(npyfile, comm=comm_self)
    np.testing.assert_allclose(loaded_data, data)


def test_same_numpy_load_transposed(comm_self, npyfile):
    data = np.random.random(size=(2, 3)).T
    np.save(npyfile, data)
    loaded_data = load_npy(npyfile, comm=comm_self)
    np.testing.assert_allclose(loaded_data, data)


def test_load_same_numpy_save(comm_self, npyfile):
    data = np.random.random(size=(2, 3))
    save_npy(npyfile, data, comm=comm_self)
    loaded_data = np.load(npyfile)
    np.testing.assert_allclose(loaded_data, data)
    assert np.isfortran(data) == np.isfortran(loaded_data)


def test_same_numpy_save_transposed(comm_self, npyfile):
    data = np.random.random(size=(2, 3)).T
    save_npy(npyfile, data, comm=comm_self)
    loaded_data = np.load(npyfile)
    np.testing.assert_allclose(loaded_data, data)
    assert np.isfortran(data) == np.isfortran(loaded_data)


# @pytest.mark.filterwarnings("error: ResourceWarnings")
# unfortunately the ResourceWarnings are never raised even when the files
# are not closed
def test_raises_and_no_resourcewarnings(comm_self):
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter(
            "always")  # deactivate hiding of ResourceWarnings

        with pytest.raises(MPIFileTypeError):
            load_npy(os.path.join(testdir, "wrongnpyfile.npy"), comm=comm_self)

        # assert no warning is a ResourceWarning
        for wi in w:
            assert not issubclass(wi.category, ResourceWarning)


def test_corrupted_file(comm_self):
    """
    tests that the reader behaves decently when trying to open a file having
    the wrong format see issue #23
    """
    # create test corrupted file

    with open("corrupted.dummy", "w") as f:
        f.write("dfgdfghlkjhgiuhdfg")

    with pytest.raises(MPIFileTypeError):
        MPIFileViewNPY("corrupted.dummy", comm=comm_self)


@pytest.mark.skipif(NuMPI._has_mpi4py,
                    reason="filestreams are not supported when "
                           "                                         NuMPI "
                           "is using with mpi4py")
def test_filestream(comm_self, npyfile):
    data = np.random.normal(size=(4, 6))

    np.save(npyfile, data)
    with open(npyfile, mode="r") as f:
        read_data = load_npy(f, comm=comm_self)
        np.testing.assert_allclose(read_data, data)
    with open(npyfile, mode="rb") as f:
        read_data = load_npy(f, comm=comm_self)
        np.testing.assert_allclose(read_data, data)


@pytest.mark.skipif(NuMPI._has_mpi4py,
                    reason="filestreams are not supported when NuMPI "
                           "is using with mpi4py")
@pytest.mark.parametrize("mode", ["r"] if NuMPI._has_mpi4py else ["r", "rb"])
def test_make_mpi_file_view(comm_self, npyfile, mode):
    data = np.random.normal(size=(4, 6))

    np.save(npyfile, data)
    with open(npyfile, mode=mode) as f:
        fileview = make_mpi_file_view(f, comm=comm_self)
        read_data = fileview.read()
        np.testing.assert_allclose(read_data, data)

        # assert data can be read several times
        read_data = fileview.read()
        np.testing.assert_allclose(read_data, data)



import pytest
import numpy as np

try:
    import scipy.optimize

    _scipy_present = True
except ModuleNotFoundError:
    _scipy_present = False

from NuMPI.Tools import Reduction
from NuMPI.Optimization.ccg_without_restart import constrained_conjugate_gradients

from test.Optimization.MPI_minimization_problems import MPI_Quadratic


def test_bugnicourt_cg(comm):
    n = 128

    obj = MPI_Quadratic(n, pnp=Reduction(comm), )

    xstart = np.random.normal(size=obj.nb_subdomain_grid_pts)

    res = constrained_conjugate_gradients(
        obj.f_grad,
        obj.hessian_product,
        x0=xstart,
        communicator=comm
    )
    assert res.success, res.message
    print(res.nit)


@pytest.mark.skipif(not _scipy_present, reason='scipy not present')
def test_bugnicourt_cg_arbitrary_bounds(comm):
    n = 128
    np.random.seed(0)
    obj = MPI_Quadratic(n, pnp=Reduction(comm), )

    xstart = np.random.normal(size=obj.nb_subdomain_grid_pts)
    bounds = np.random.normal(size=obj.nb_subdomain_grid_pts)

    res = constrained_conjugate_gradients(
        obj.f_grad,
        obj.hessian_product,
        x0=xstart,
        communicator=comm,
        bounds=bounds,
        gtol=1e-8
    )
    assert res.success, res.message

    assert (res.x >= bounds).all()
    print(res.nit)

    # TODO: we are not checking yet that the result is the same in parallel.
    if comm.size == 1:
        bnds = tuple([(b, None) for b in bounds])

        res_ref = scipy.optimize.minimize(
            obj.f_grad,
            x0=xstart, bounds=bnds, method="L-BFGS-B", jac=True,
            options=dict(gtol=1e-8, ftol=0))
        assert res_ref.success, res_ref.message

        np.testing.assert_allclose(res.x, res_ref.x, atol=1e-6)


def test_bugnicourt_cg_active_bounds(comm):
    n = 128
    np.random.seed(0)
    obj = MPI_Quadratic(n, pnp=Reduction(comm), xmin=np.random.normal(size=n))

    xstart = np.random.normal(size=obj.nb_subdomain_grid_pts)

    res = constrained_conjugate_gradients(
        obj.f_grad,
        obj.hessian_product,
        x0=xstart,
        communicator=comm
    )
    assert res.success, res.message
    print(res.nit)
    print(np.count_nonzero(res.x == 0))


def test_bugnicourt_cg_mean_val(comm):
    n = 128

    obj = MPI_Quadratic(n, pnp=Reduction(comm), )

    xstart = np.random.normal(size=obj.nb_subdomain_grid_pts)

    res = constrained_conjugate_gradients(
        obj.f_grad,
        obj.hessian_product,
        x0=xstart,
        mean_val=1.,
        communicator=comm
    )
    assert res.success, res.message
    print(res.nit)


def test_bugnicourt_cg_mean_val_active_bounds(comm):
    n = 128
    np.random.seed(0)
    obj = MPI_Quadratic(n, pnp=Reduction(comm), xmin=np.random.normal(size=n))

    xstart = np.random.normal(size=obj.nb_subdomain_grid_pts)

    res = constrained_conjugate_gradients(
        obj.f_grad,
        obj.hessian_product,
        x0=xstart,
        mean_val=1.,
        communicator=comm
    )
    assert res.success, res.message
    print(res.nit)



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import pytest

import numpy as np
from NuMPI.Tools import Reduction
from test.Optimization.MPI_minimization_problems import (
    MPI_Objective_Interface, MPI_Quadratic
    )

import time
import test.Optimization.minimization_problems as mp

from NuMPI.Optimization.MPI_LBFGS_Matrix_H import LBFGS


def timer(fun, *args, **kwargs):
    start = time.perf_counter()
    res = fun(*args, **kwargs)
    delay = time.perf_counter() - start
    return res, delay


def test_linesearch():
    pass


@pytest.mark.parametrize("n", [10, 20, 50])
def test_quadratic_analytical_min(comm, n):
    """
    Compares the result with the analyticaly known posistion of the minimum
    :return:
    """

    def printMPI(msg):
        for i in range(comm.Get_size()):
            comm.barrier()
            if comm.Get_rank() == i:
                print("Proc {}: {}".format(i, msg))

    PObjective = MPI_Quadratic(nb_domain_grid_pts=n, pnp=Reduction(comm))

    x0 = PObjective.startpoint()

    res = LBFGS(PObjective.f_grad, x0, jac=True, maxcor=5, maxiter=100,
                gtol=1e-14, ftol=0, pnp=Reduction(comm))
    #                        ^ only terminates if gradient condition is
    #                        satisfied
    assert res.success
    assert res.message == "CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL"
    np.testing.assert_allclose(res.x, PObjective.xmin(), atol=1e-13)

    assert np.abs(res.fun - PObjective.minVal(n)) < 1e-7


@pytest.mark.parametrize("n", [10, 20, 50])
def test_rosenbrock_analytical_min(comm, n):
    """
    Compares the result with the analyticaly known posistion of the minimum
    :return:
    """

    def printMPI(msg):
        for i in range(comm.Get_size()):
            comm.barrier()
            if comm.Get_rank() == i:
                print("Proc {}: {}".format(i, msg))

    Objective = mp.Extended_Rosenbrock

    PObjective = MPI_Objective_Interface(Objective, nb_domain_grid_pts=n,
                                         comm=comm)

    x0 = PObjective.startpoint()

    res = LBFGS(PObjective.f_grad, x0, jac=True, maxcor=5, maxiter=100,
                gtol=1e-12, ftol=0, pnp=Reduction(comm))
    #                        ^ only terminates if gradient condition is
    #                        satisfied
    assert res.success
    assert res.message == "CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL"
    np.testing.assert_allclose(res.x, PObjective.xmin(), atol=1e-16, rtol=1e-7)

    assert np.abs(res.fun - Objective.minVal(n)) < 1e-7


# def test_ftol(comm):
#    pass

# def test_gtol(comm):
#    pass

# def test_g2tol(comm):
#    pass

# def test_alltol(comm):
#    pass

@pytest.mark.skip(reason="just plotting")
def test_time_complexity(comm):
    maxcor = 5
    Objective = mp.Extended_Rosenbrock
    n = np.array([10, 100, 1000, 1e4, 1e5, 1e6], dtype=int)
    t = np.zeros(len(n), dtype=float)
    res = [None] * len(n)
    pnp = Reduction(comm)
    for i in range(len(n)):
        PObjective = MPI_Objective_Interface(Objective,
                                             nb_domain_grid_pts=n[i],
                                             comm=comm)
        x0 = PObjective.startpoint()

        res[i], t[i] = timer(LBFGS, PObjective.f_grad, x0, jac=True,
                             maxcor=maxcor, maxiter=100000, gtol=(1e-5),
                             pnp=pnp)

        assert res[i].success

    if True:
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.plot(n, t / n, '+-', label="time / DOF")
        ax.plot(n, [t[i] / n[i] / res[i].nit for i in range(len(n))], '+-',
                label="time per DOF per iteration")
        ax2 = plt.twinx(ax)
        ax2.plot(n, [res[i].nit for i in range(len(n))], 'o', label="nit")
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_ylabel("t/n (s)")
        ax.set_xlabel("DOF")
        ax.legend()
        ax2.legend()
        # ax.plot(n,n,c='gray')
        plt.show(block=True)



#
# Copyright 2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# This was migrated from muSpectre byy Antoine Sanner. Lars Pastewka is author

import numpy as np
import pytest

try:
    # raise ImportError()
    from netCDF4 import Dataset

    _has_netCDF4 = True
except ImportError:
    _has_netCDF4 = False
###

if _has_netCDF4:
    from NuMPI.IO.NetCDF import NCStructuredGrid
else:
    pytestmark = pytest.mark.skip("skip because no NetCDF installed")


class DomainDecomposition():
    def __init__(self, nb_domain_grid_pts, communicator):
        nprocs = communicator.Get_size()
        rank = communicator.Get_rank()

        self.subdomain_slices = [slice(None), ] * len(nb_domain_grid_pts)
        self.subdomain_locations = [0, ] * len(nb_domain_grid_pts)
        self.nb_subdomain_grid_pts = list(nb_domain_grid_pts)

        step = nb_domain_grid_pts[0] // nprocs
        # decomposition along first axis
        if rank == nprocs - 1:
            self.subdomain_slices[0] = slice(rank * step, None)
            self.subdomain_locations[0] = rank * step
            self.nb_subdomain_grid_pts[0] = nb_domain_grid_pts[0] - rank * step
        else:
            self.subdomain_slices[0] = slice(rank * step, (rank + 1) * step)
            self.subdomain_locations[0] = rank * step
            self.nb_subdomain_grid_pts[0] = step


@pytest.fixture(params=[(11,), (11, 23), (11, 23, 7)])
def self(request, comm):
    # This fixture mimics the setup in the unit test class.
    # This test was written in the unittest syntax before.

    class self:
        pass

    self.nb_grid_pts = request.param
    self.tensor_shape = tuple(list(self.nb_grid_pts) + [3, 3])
    self.scalar_grid = np.arange(np.prod(self.nb_grid_pts)).reshape(self.nb_grid_pts)
    self.tensor_grid = np.arange(np.prod(self.tensor_shape)).reshape(self.tensor_shape)

    self.communicator = comm

    self.domain_decomposition = DomainDecomposition(self.nb_grid_pts, self.communicator)

    return self


def test_write_read_domain(self):
    if self.communicator is None:
        nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='w',
                              nb_domain_grid_pts=self.nb_grid_pts)
    else:
        nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='w',
                              nb_domain_grid_pts=self.nb_grid_pts,
                              decomposition='domain',
                              subdomain_locations=self.domain_decomposition.subdomain_locations,
                              nb_subdomain_grid_pts=self.domain_decomposition.nb_subdomain_grid_pts,
                              communicator=self.communicator)
    nc.scalar = self.scalar_grid
    nc.tensor = self.tensor_grid
    nc[3].per_frame_tensor = self.tensor_grid
    nc.close()
    self.communicator.barrier()
    # Check that the file structure is correct
    nc = Dataset('test_{}d.nc'.format(len(self.nb_grid_pts)), 'r', comm=None)
    dimensions = ['frame', 'grid_x', 'tensor_3']
    if len(self.nb_grid_pts) > 1:
        dimensions += ['grid_y']
    if len(self.nb_grid_pts) == 3:
        dimensions += ['grid_z']
    assert set(nc.dimensions) == set(dimensions)
    assert len(nc.dimensions['frame']) == 4
    assert len(nc.dimensions['grid_x']) == self.nb_grid_pts[0]
    if len(self.nb_grid_pts) > 1:
        assert len(nc.dimensions['grid_y']) == self.nb_grid_pts[1]
    if len(self.nb_grid_pts) == 3:
        assert len(nc.dimensions['grid_z']) == self.nb_grid_pts[2]
    assert len(nc.dimensions['tensor_3']) == 3
    nc.close()

    # Read file and check data
    nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='r', communicator=None)
    assert np.equal(tuple(nc.nb_domain_grid_pts), tuple(self.nb_grid_pts)).all()
    assert np.equal(nc.scalar, self.scalar_grid).all()
    assert np.equal(nc.tensor, self.tensor_grid).all()
    assert np.equal(nc[3].per_frame_tensor, self.tensor_grid).all()
    nc.close()


def test_write_read_subdomain(self):
    scalar_grid = self.scalar_grid[tuple(self.domain_decomposition.subdomain_slices)]
    tensor_grid = self.tensor_grid[tuple(self.domain_decomposition.subdomain_slices)]

    if self.communicator is None:
        nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='w',
                              nb_domain_grid_pts=self.nb_grid_pts)
    else:
        nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='w',
                              nb_domain_grid_pts=self.nb_grid_pts,
                              decomposition='subdomain',
                              subdomain_locations=self.domain_decomposition.subdomain_locations,
                              nb_subdomain_grid_pts=self.domain_decomposition.nb_subdomain_grid_pts,
                              communicator=self.communicator)
    nc.scalar = scalar_grid
    nc.tensor = tensor_grid
    nc[3].per_frame_tensor = tensor_grid
    nc.close()
    self.communicator.barrier()
    # Check that the file structure is correct
    nc = Dataset('test_{}d.nc'.format(len(self.nb_grid_pts)), 'r', comm=None)
    dimensions = ['frame', 'grid_x', 'tensor_3']
    if len(self.nb_grid_pts) > 1:
        dimensions += ['grid_y']
    if len(self.nb_grid_pts) == 3:
        dimensions += ['grid_z']
    assert np.equal(set(nc.dimensions), set(dimensions))
    assert len(nc.dimensions['frame']) == 4
    assert len(nc.dimensions['grid_x']) == self.nb_grid_pts[0]
    if len(self.nb_grid_pts) > 1:
        assert len(nc.dimensions['grid_y']) == self.nb_grid_pts[1]
    if len(self.nb_grid_pts) == 3:
        assert len(nc.dimensions['grid_z']) == self.nb_grid_pts[2]
    assert len(nc.dimensions['tensor_3']) == 3
    nc.close()

    # Read file and check data
    nc = NCStructuredGrid('test_{}d.nc'.format(len(self.nb_grid_pts)), mode='r', communicator=None)
    assert np.equal(tuple(nc.nb_domain_grid_pts), tuple(self.nb_grid_pts)).all()
    assert np.equal(nc.scalar, self.scalar_grid).all()
    assert np.equal(nc.tensor, self.tensor_grid).all()
    assert np.equal(nc[3].per_frame_tensor, self.tensor_grid).all()
    nc.close()



#
# Copyright 2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
from NuMPI.Tools.Reduction import Reduction
from test.Optimization.MPI_minimization_problems import MPI_Quadratic
import time

from NuMPI.Optimization.MPI_LBFGS_Matrix_H import LBFGS

from NuMPI import MPI
import cProfile


def profile(filename=None, comm=MPI.COMM_WORLD):
    def prof_decorator(f):
        def wrap_f(*args, **kwargs):
            pr = cProfile.Profile()
            pr.enable()
            result = f(*args, **kwargs)
            pr.disable()

            if filename is None:
                pr.print_stats()
            else:
                filename_r = filename + ".{}".format(comm.rank)
                pr.dump_stats(filename_r)

            return result

        return wrap_f

    return prof_decorator


def timer(fun, *args, **kwargs):
    start = time.perf_counter()
    res = fun(*args, **kwargs)
    delay = time.perf_counter() - start
    return res, delay


np.random.seed(1)
n = int(1e7)

# Objective = mp.Extended_Rosenbrock
maxcor = 10
factors = 0.1 + np.random.random(n)
startpoint = np.random.normal(size=n)

comm = MPI.COMM_WORLD

pnp = Reduction(comm)
# PObjective = MPI_Objective_Interface(Objective, nb_domain_grid_pts=n, comm=comm)
PObjective = MPI_Quadratic(nb_domain_grid_pts=n, pnp=pnp, factors=factors,
                           startpoint=startpoint)
x0 = PObjective.startpoint()

LBFGS = profile("profile_out", comm)(LBFGS)
res, t = timer(LBFGS, PObjective.f, x0, jac=PObjective.grad, maxcor=maxcor,
               maxiter=100000, gtol=(1e-5), pnp=pnp)
assert res.success
if MPI.COMM_WORLD.Get_rank() == 0:
    print("elapsed time: {}".format(t))

if comm.rank == 0:
    print("to vizualize the profile, execute `snakeviz profile_out.{rank}`")



#
# Copyright 2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
This runs LBFGS and scipy L-BFGS-B in serial and compare the number of iterations needed
"""

import os
from NuMPI.Optimization import LBFGS
from test.Optimization.minimization_problems import Trigonometric, Extended_Rosenbrock
import numpy as np
import scipy

import matplotlib.pyplot as plt

# import sys
n = 10 ** 3


class decorated_objective:
    def __init__(self, objective):
        self.objective = objective
        self.neval = 0
        self.energies = []
        self.maxgradients = []

    def __call__(self, *args, **kwargs):
        f, grad = self.objective(*args, **kwargs)
        self.neval += 1
        self.energies.append(f)
        self.maxgradients.append(np.max(abs(grad)))
        return f, grad


for Objective in [Trigonometric, Extended_Rosenbrock]:
    fig, (axEn, axgrad) = plt.subplots(2, 1, sharex=True)
    for method, name, c in zip([LBFGS, "L-BFGS-B"], ["NuMPI", "scipy"], ["C1", "C0"]):
        objective_monitor = decorated_objective(Objective.f_grad)
        result = scipy.optimize.minimize(objective_monitor, Objective.startpoint(n),
                                         jac=True, method=method,
                                         options={"gtol": 1e-6, "ftol": 1e-20, "maxcor": 20})
        assert result.success

        axgrad.plot(range(objective_monitor.neval),
                    objective_monitor.maxgradients, label="{}".format(name))
        axEn.plot(
            range(objective_monitor.neval),
            (np.array(objective_monitor.energies) - objective_monitor.energies[-1]) / (
                    objective_monitor.energies[0] - objective_monitor.energies[-1]),
            label="{}".format(name), c=c)

        print("{}, {}: nevals {}, nit {}, evals/it {}"
              .format(name,
                      Objective.__name__,
                      objective_monitor.neval,
                      result.nit, float(objective_monitor.neval) / result.nit))

    axEn.set_xlabel("# of objective evaluations")
    axEn.set_ylabel("E(i)-E(last) / (E(0)-E(last))")
    axEn.set_yscale("log")

    axgrad.set_yscale("log")
    axgrad.set_ylabel(r"$|grad|_{\infty}$")
    axgrad.legend()

    for a in (axEn, axgrad):
        a.set_xlabel("# of objective evaluations")
        a.label_outer()

    fig.suptitle("{}, n={}".format(Objective.__name__, n))
    fig.savefig("{}_{}.png".format(os.path.basename(__file__), Objective.__name__))



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Tests the that the objective functions provided for the tests are
correctly implemented
"""

import numpy as np
import pytest
from NuMPI import MPI

try:
    import scipy.optimize

    _scipy_present = True
except ModuleNotFoundError:
    _scipy_present = False

import test.Optimization.minimization_problems as mp
from test.Optimization.MPI_minimization_problems import MPI_Objective_Interface


@pytest.mark.skipif(
    MPI.COMM_WORLD.Get_size() > 1,
    reason="tests only serial funcionalities, please execute with pytest")
@pytest.mark.parametrize("Objective",
                         [mp.Trigonometric, mp.Extended_Rosenbrock])
@pytest.mark.parametrize("n", [2, 4, 10])
def test_directional_derivative(Objective, n):
    """
    Asserts the nmerical and the Gradient provided by Objective are close
    :param Objective: which function to minimize
    :param n: number of dimensions
    :return:
    """
    _verbose = False
    if _verbose:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.set_xscale("log")
        ax.set_yscale("log")

    for i in range(12):
        x = Objective.bounds[0] + (Objective.bounds[1] - Objective.bounds[
            0]) * np.random.random(n)
        x.shape = (-1, 1)
        u = np.random.normal(size=n)  # Direction
        u /= np.linalg.norm(u, 2)  # normalize
        u.shape = (-1, 1)

        # TODO: Don't know which Tolerance to associate with which eps
        epsilons = np.array([1e-3, 1e-5, 1e-6])
        # der_numerical = np.zeros_like(epsilons)
        # der_analytical = np.zeros_like(epsilons)
        errorratios = np.zeros_like(epsilons)
        for eps, i in zip(epsilons, range(len(epsilons))):
            # print((Objective.f(x + u * eps) - Objective.f(x)) / eps)
            der_numerical = (Objective.f(x + u * eps) - Objective.f(x)) / eps
            der_analytical = (Objective.grad(x).T @ u).item()

            errorratios[i] = np.abs(der_numerical - der_analytical) / eps
        errorratios /= errorratios[0]
        if _verbose:
            ax.plot(epsilons, errorratios)

        assert (errorratios < 10).all(), "error_ratios {}".format(
            abs(der_numerical - der_analytical) / der_analytical)
    if _verbose:
        plt.show(block=True)


@pytest.mark.skipif(not _scipy_present, reason='scipy not present')
@pytest.mark.skipif(
    MPI.COMM_WORLD.Get_size() > 1,
    reason="tests only serial funcionalities, please execute with pytest")
@pytest.mark.parametrize("Objective",
                         [mp.Extended_Rosenbrock, mp.Trigonometric])
@pytest.mark.parametrize("n", [2, 4, 10, 20])
def test_Gradient(Objective, n):
    """
    Asserts the numerical Gradient converges with order of the step to the
    analyitical one.

    This uses check_gradient from scipy

    :param Objective: which function to minimize
    :param n: number of dimensions
    :return:
    """
    _verbose = False
    if _verbose:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.set_xscale("log")
        ax.set_yscale("log")

    for i in range(12):  # reproduce to have a bit of statistics
        x = Objective.bounds[0] + (Objective.bounds[1] - Objective.bounds[
            0]) * np.random.random(n)

        # x.shape = (-1,1)

        epsilons = np.array([1e-3, 1e-5, 1e-6])
        errors = np.array([scipy.optimize.check_grad(Objective.f,
                                                     Objective.grad, x,
                                                     epsilon=eps) for eps in
                           epsilons])  # should be O(epsilon)
        errorratios = errors / epsilons  # should be O(1)
        errorratios /= errorratios[0]
        if _verbose:
            ax.plot(epsilons, errorratios)

        # print(errorratios)
        assert np.prod(errorratios < 10), "errorratios = {}".format(
            errorratios)
    if _verbose:
        plt.show(block=True)


@pytest.mark.skipif(
    MPI.COMM_WORLD.Get_size() > 1,
    reason="tests only serial funcionalities, please execute with pytest")
def test_MPI_Parallel_Interface(comm):
    """ Tests the parallel wrapping of objectives
    Test if parallel Version gives the same as the serial Version
    """

    def printMPI(msg):
        for i in range(comm.Get_size()):
            comm.barrier()
            if comm.Get_rank() == i:
                print("Proc {}: {}".format(i, msg))

    n = 10

    par = MPI_Objective_Interface(mp.Extended_Rosenbrock, nb_domain_grid_pts=n,
                                  comm=comm)

    printMPI(par.counts)

    # ref = mp.Extended_Rosenbrock

    np.testing.assert_array_equal(
        mp.Extended_Rosenbrock.startpoint(n)[par.subdomain_slices],
        par.startpoint())
    np.testing.assert_almost_equal(
        mp.Extended_Rosenbrock.f(mp.Extended_Rosenbrock.startpoint(n)),
        par.f(par.startpoint()),
        err_msg="Different Function Value at startpoint")
    np.testing.assert_allclose(
        mp.Extended_Rosenbrock.grad(mp.Extended_Rosenbrock.startpoint(n))[
            par.subdomain_slices],
        par.grad(par.startpoint()),
        err_msg="Different Gradient Value at startpoint")



"""
Constrained Conjugate Gradient for moderately nonlinear problems.
This implementation is mainly based upon Bugnicourt et. al. - 2018, OUT_LIN
algorithm.
"""

import numpy as np

from inspect import signature

from ..Tools import Reduction
from .result import OptimizeResult


def constrained_conjugate_gradients(fun, hessp,
                                    x0, mean_val=None,
                                    gtol=1e-8,
                                    maxiter=3000,
                                    callback=None,
                                    communicator=None,
                                    bounds=None
                                    ):
    '''
    Constrained conjugate gradient algorithm from Bugnicourt et al. [1].

    Parameters
    __________

    fun :   callable.
                The objective function to be minimized.
                            fun(x) -> float(energy),ndarray(gradient)
                where x is the input ndarray. The energy is actually never used.

    hessp : callable
            Function to evaluate the hessian product of the objective.
            Hessp should accept either 1 argument (descent direction) or
            2 arguments (x,descent direction).
                            hessp(des_dir)->ndarray
                                    or
                            hessp(x,des_dir)->ndarray
            where x is the input ndarray and des_dir is the descent direction.

    x0 : ndarray
         Initial guess.

    gtol : float, optional
           Default value : 1e-8
           convergence criterion is max(abs) and norm2 of the projected
           gradient < gtol.
    mean_value : int/float, optional
               If you want to apply the mean_value constraint then provide an
               int/float value to the mean_value.

    residual_plot : bool, optional
                    Generates a plot between the residual and iterations.

    maxiter : int, optional
              Default, maxiter=5000
              Maximum number of iterations after which the program will exit.

    Returns
    -------
    OptimizeResult  : scipy.optimize object.
        Attributes:
         success: bool
         x: x,
         jac: residual = gradient(x),
         nit: n_iterations,
         message: 'CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL' or 'NO CONVERGENCE: MAXITERATIONS REACHED'

    References
    __________

    ..[1]   :   Bugnicourt, Romain & Sainsot, Philippe & Dureisseix, David &
                Gauthier, Catherine & Lubrecht, Ton. (2018). FFT-Based Methods
                for Solving a Rough Adhesive Contact: Description and
                Convergence Study.
                Tribology Letters. 66. 10.1007/s11249-017-0980-z.

    ..[2]   :   Vollebregt, E. A. H. J Optim Theory Appl 162, 931–953 (2014)
                The Bound-Constrained Conjugate Gradient Method for Non-negative Matrices

    '''
    if communicator is None:
        comm = np
        nb_DOF = x0.size
    else:
        comm = Reduction(communicator)
        nb_DOF = comm.sum(x0.size)

    x = x0.copy()
    x = x.flatten()

    if bounds is None:
        bounds = np.zeros_like(x)

    mask_bounds = bounds > - np.infty
    nb_bounds = comm.sum(np.count_nonzero(mask_bounds))
    mean_bounds = comm.sum(bounds) / nb_bounds

    if mean_val is not None and nb_bounds < nb_DOF:
        raise ValueError("mean_value constrained mode not compatible "
                         "with partially bound system")
        # There are ambiguities on how to compute the mean values

    '''Initial Residual = A^(-1).(U) - d A −1 .(U ) −  ∂ψadh/∂g'''
    residual = fun(x)[1]

    mask_neg = x <= bounds
    x[mask_neg] = bounds[mask_neg]

    if mean_val is not None:
        #
        mask_nonzero = x > bounds
        N_mask_nonzero = comm.sum(np.count_nonzero(mask_nonzero))
        residual = residual - comm.sum(residual[mask_nonzero]) / N_mask_nonzero

    '''Apply the admissible Lagrange multipliers.'''
    mask_res = residual >= 0
    mask_bounded = np.logical_and(mask_neg, mask_res)
    residual[mask_bounded] = 0.0

    if mean_val is not None:
        #
        mask_nonzero = residual != 0
        N_nonzero = comm.sum(np.count_nonzero(mask_nonzero))
        residual[mask_nonzero] = residual[mask_nonzero] - comm.sum(
            residual[mask_nonzero]) / N_nonzero
    '''INITIAL DESCENT DIRECTION'''
    des_dir = -residual

    n_iterations = 1

    for i in range(1, maxiter + 1):
        sig = signature(hessp)
        if len(sig.parameters) == 2:
            hessp_val = hessp(x, des_dir)
        elif len(sig.parameters) == 1:
            hessp_val = hessp(des_dir)
        else:
            raise ValueError('hessp function has to take max 1 arg (descent '
                             'dir) or 2 args (x, descent direction)')
        denominator_temp = comm.sum(des_dir.T * hessp_val)
        # Here we could evaluate this directly in Fourier space (Parseval)
        # and spare one FFT.
        # See issue #47

        if denominator_temp == 0:
            print("it {}: denominator for alpha is 0".format(i))

        alpha = -comm.sum(residual.T * des_dir) / denominator_temp

        if alpha < 0:
            print("it {} : hessian is negative along the descent direction. "
                  "You will probably need linesearch "
                  "or trust region".format(i))

        x += alpha * des_dir

        '''finding new contact points and making the new_gap admissible
        according to these contact points.'''

        mask_neg = x <= bounds
        x[mask_neg] = bounds[mask_neg]

        if mean_val is not None:
            # x = (mean_val / comm.sum(x) * nb_DOF) * x
            # below is just a more complicated version of this compatible with
            # more general bounds
            x = bounds + (mean_val - mean_bounds) \
                / (comm.sum(x) / nb_DOF - mean_bounds) * (x - bounds)
        residual_old = residual

        '''
        In Bugnicourt's paper
        Residual = A^(-1).(U) - d A −1 .(U ) −  ∂ψadh/∂g
        '''
        residual = fun(x)[1]

        if mean_val is not None:
            mask_nonzero = x > bounds
            N_mask_nonzero = comm.sum(np.count_nonzero(mask_nonzero))
            residual = residual - comm.sum(
                residual[mask_nonzero]) / N_mask_nonzero

        '''Apply the admissible Lagrange multipliers.'''
        mask_res = residual >= 0
        mask_bounded = np.logical_and(mask_neg, mask_res)
        residual[mask_bounded] = 0.0

        if mean_val is not None:
            mask_nonzero = residual != 0
            N_nonzero = comm.sum(np.count_nonzero(mask_nonzero))
            residual[mask_nonzero] = residual[mask_nonzero] - comm.sum(
                residual[mask_nonzero]) / N_nonzero
            # assert np.mean(residual) < 1e-14 * np.max(abs(residual))

        '''Computing beta for updating descent direction
            In Bugnicourt's paper:
            beta = num / denom
            num = new_residual_transpose . (new_residual - old_residual)
            denom = alpha * descent_dir_transpose . (A_inverse - d2_ψadh).
            descent_dir '''

        # beta = np.sum(residual.T * hessp_val) / denominator_temp
        beta = comm.sum(residual * (residual - residual_old)) / (
                alpha * denominator_temp)

        des_dir_old = des_dir
        des_dir = -residual + beta * des_dir_old

        des_dir[mask_bounded] = 0

        if callback:
            callback(x)

        n_iterations += 1

        if comm.max(abs(residual)) <= gtol:
            result = OptimizeResult(
                {
                    'success': True,
                    'x': x,
                    'jac': residual,
                    'nit': i,
                    'message': 'CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL',
                    })
            return result

        elif i == maxiter - 1:
            result = OptimizeResult(
                {
                    'success': False,
                    'x': x,
                    'jac': residual,
                    'nit': i,
                    'message': 'NO CONVERGENCE: MAXITERATIONS REACHED'
                    })

            return result



#
# Copyright 2018 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


from matplotlib.patches import FancyArrowPatch
import numpy as np

rec = 0


def asscalar(a):
    #    global rec
    #    print(rec)
    #    rec=rec+1;
    #    print(a)
    if isinstance(a, np.ndarray):
        return asscalar(a.tolist()[0])
    elif isinstance(a, (list, tuple)):
        return asscalar(a[0])
    else:
        return a


def draw_npArrow2D(ax, start, end=None, delta=None, arrowprops={}, label=None, textprops={}):
    arrow_prop_dict = dict(mutation_scale=10, arrowstyle='-|>', color='k', shrinkA=0, shrinkB=0)
    for key in arrowprops:
        arrow_prop_dict[key] = arrowprops[key]

    xs = asscalar(start[0])
    ys = asscalar(start[1])

    if end is not None:
        xe = asscalar(end[0])
        ye = asscalar(end[1])
    elif delta is not None:
        xe = xs + asscalar(delta[0])
        ye = ys + asscalar(delta[1])
    else:
        raise AssertionError('Missing Argument end or delta')

    ax.add_artist(FancyArrowPatch((xs, ys), (xe, ye), **arrow_prop_dict))
    if label is not None:
        ax.text(xe, ye, label, **textprops)



import sys

import netCDF4

mpi_required = sys.argv[1] in ["yes", "1"]

print('NetCDF configuration:')
print('CDF5 {}'.format(netCDF4.__has_cdf5_format__))
print('Parallel4: {}'.format(netCDF4.__has_parallel4_support__))
print('PnetCDF: {}'.format(netCDF4.__has_pnetcdf_support__))

# Make sure that we have CDF5 support
assert netCDF4.__has_cdf5_format__

if mpi_required:
    # Make sure that we have the parallel version running
    # assert netCDF4.__has_parallel4_support__
    assert netCDF4.__has_pnetcdf_support__



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np

from NuMPI import MPI
from NuMPI.Tools import Reduction


class MPI_Extended_Rosenbrock():  # TODO: This doesn't work
    """

    This is the Definition like in Moré et al. and not like in scipy

    n should be even

    :param x: 1d array
    :return:
    """
    bounds = (-4, 4)

    def __init__(self, nb_domain_grid_pts, pnp=Reduction()):
        raise NotImplementedError("Need to implement communication")
        comm = pnp.comm
        nprocs = comm.Get_size()
        rank = comm.Get_rank()

        step = nb_domain_grid_pts // nprocs

        if rank == nprocs - 1:
            self.subdomain_slices = slice(rank * step, None)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = nb_domain_grid_pts - rank * step
        else:
            self.subdomain_slices = slice(rank * step, (rank + 1) * step)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = step

        # helps to select the data that has odd or even index in the global
        # array
        self._sl_odd = slice(self.subdomain_locations % 2, None, 2)
        self._sl_even = slice((self.subdomain_locations + 1) % 2, None, 2)
        self.pnp = pnp
        self.nfeval = 0
        self.ngradeval = 0

    def f_grad(self, x):
        self.nfeval += 1
        self.ngradeval += 1

        x_odd = x[self._sl_odd]
        x_even = x[self._sl_even]

        sumf2 = (self.pnp.sum(
            100 * (x_odd - x_even ** 2) ** 2 + (1 - x_even) ** 2)).item()

        grad = np.zeros_like(x)
        grad[self._sl_odd] = 200 * (x_odd - x_even ** 2)
        grad[self._sl_even] = -400 * x_even * (x_odd - x_even ** 2) - 2 * (
                    1 - x_even)  # # d / dx2l-1

        return sumf2, grad

    def f(self, x):
        self.ngradeval -= 1
        return self.f_grad(x)[0]

    def grad(self, x):
        self.nfeval -= 1  # compensate for self.nfeval +=1 in f_grad
        return self.f_grad(x)[1]

    def startpoint(self):
        """
        standard starting point
        :param n:
        :return: array of shape (1,n)
        """
        x0 = np.zeros(self.nb_subdomain_grid_pts, dtype=float)
        x0.shape = (-1, 1)

        x0[self._sl_even] = -1.2
        x0[self._sl_odd] = 1

        return x0

    @staticmethod
    def minVal(*args):
        return 0

    def xmin(self):
        """
        Location of minimum according to

        Mori, J. J., Garbow, B. S. & Hillstrom, K. E. Testing Unconstrained
        Optimization Software. 25 (1981).

        This function not necessarily have only one Minimum in higher
        dimensional Space: see e.g. 10.1162/evco.2006.14.1.119

        :param n: number of DOF
        :return: array of size n
        """

        return np.ones((self.nb_subdomain_grid_pts, 1), dtype=float)


class MPI_Quadratic():
    """
    n should be even

    :param x: 1d array
    :return:
    """
    bounds = (-4, 4)

    def __init__(self, nb_domain_grid_pts, pnp=Reduction(), factors=None,
                 startpoint=None, xmin=None):

        comm = pnp.comm
        nprocs = comm.Get_size()
        rank = comm.Get_rank()

        step = nb_domain_grid_pts // nprocs

        if rank == nprocs - 1:
            self.subdomain_slices = slice(rank * step, None)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = nb_domain_grid_pts - rank * step
        else:
            self.subdomain_slices = slice(rank * step, (rank + 1) * step)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = step

        # helps to select the data that has odd or even index in the global
        # array
        self.pnp = pnp

        if xmin is None:
            self._xmin = np.zeros(self.nb_subdomain_grid_pts)
        else:
            self._xmin = xmin[self.subdomain_slices]

        if factors is not None:
            self.factors = factors[self.subdomain_slices]
        else:
            self.factors = np.random.random(self.nb_subdomain_grid_pts) + 0.1

        if startpoint is not None:
            self._startpoint = startpoint[self.subdomain_slices]
        else:
            self._startpoint = np.random.normal(
                size=self.nb_subdomain_grid_pts)

        self.nfeval = 0
        self.ngradeval = 0

    def f_grad(self, x):
        self.nfeval += 1
        self.ngradeval += 1
        factdotx = self.factors.reshape(x.shape) \
            * (x - self._xmin.reshape(x.shape))
        return self.pnp.sum(factdotx ** 2, axis=0).item(), 2 * factdotx

    def f(self, x):
        self.nfeval += 1
        factdotx = self.factors.reshape(x.shape) \
            * (x - self._xmin.reshape(x.shape))
        return self.pnp.sum(factdotx ** 2, axis=0).item()

    def grad(self, x):
        self.ngradeval += 1
        return 2 * self.factors.reshape(x.shape) * (x - self._xmin)

    def hessian_product(self, p):
        return 2 * self.factors.reshape(p.shape) * p

    def startpoint(self):
        """
        standard starting point
        :param n:
        :return: array of shape (1,n)
        """
        return self._startpoint

    @staticmethod
    def minVal(*args):
        return 0

    def xmin(self):
        """
        Location of minimum according to
        :param n: number of DOF
        :return: array of size n
        """
        return self._xmin


class MPI_Objective_Interface():
    """
    creates an interface for an objective function computed serially to
    appear like parallel.

    That means the gradient is not the full vector but only the components
    corresponding to the subdomain

    """

    def __init__(self, Objective, nb_domain_grid_pts, comm=MPI.COMM_WORLD):

        # define the partition between the processors

        self.comm = comm
        nprocs = comm.Get_size()
        rank = comm.Get_rank()

        step = nb_domain_grid_pts // nprocs

        if rank == nprocs - 1:
            self.subdomain_slices = slice(rank * step, None)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = nb_domain_grid_pts - rank * step
        else:
            self.subdomain_slices = slice(rank * step, (rank + 1) * step)
            self.subdomain_locations = rank * step
            self.nb_subdomain_grid_pts = step

        self.counts = [step] * nprocs
        self.counts[-1] = nb_domain_grid_pts - (nprocs - 1) * step

        self.Objective = Objective
        self.nb_domain_grid_pts = nb_domain_grid_pts

    def f_grad(self, x):
        x_ = x.reshape(-1)
        fullx = np.zeros(self.nb_domain_grid_pts, dtype=x.dtype)
        self.comm.Allgatherv(x_, [fullx, self.counts])
        f, fullgrad = self.Objective.f_grad(fullx)
        return f, fullgrad[self.subdomain_slices].reshape(x.shape)

    def f(self, x):
        return self.f_grad(x)[0]

    def grad(self, x):
        return self.f_grad(x)[1]

    def startpoint(self):
        return self.Objective.startpoint(self.nb_domain_grid_pts)[
            self.subdomain_slices]

    def xmin(self):
        return self.Objective.xmin(self.nb_domain_grid_pts)[
            self.subdomain_slices]



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
from mpi4py import MPI
from NuMPI.Tools import Reduction
from test.Optimization.MPI_minimization_problems import MPI_Quadratic
import time

from NuMPI.Optimization import LBFGS


def timer(fun, *args, **kwargs):
    start = time.perf_counter()
    res = fun(*args, **kwargs)
    delay = time.perf_counter() - start
    return res, delay


def show_parallel_speedup():
    msg = ""

    orsizes = np.array([4, 8, 10, 20])
    orsizes = orsizes[orsizes <= MPI.COMM_WORLD.size]
    sizes = orsizes.copy()

    toPlot = MPI.COMM_WORLD.Get_rank() == 0 and True
    if toPlot:
        import matplotlib
        matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        fig, (ax, ax2) = plt.subplots(2, 1, sharex=True)
        ax.set_xlabel("nprocs")
        ax.set_ylabel("t[1proc] / t")
        ax2.set_xlabel("nprocs")
        ax2.set_ylabel("t (s)")
        ax2.set_yscale('log')

        ax.plot(sizes, sizes, '--k', label="ideal")

    # for n in [int(1e5),int(2e5),int(1e6),int(2e6),int(1e7)]:
    for n in [int(1e6)]:
        # sizes = orsizes[orsizes > n / 1e4]

        if len(sizes) == 0:
            continue
        t = np.zeros(len(sizes), dtype=float)
        res = [None] * len(sizes)

        # Objective = mp.Extended_Rosenbrock
        maxcor = 5
        factors = 0.1 + np.random.random(n)
        startpoint = np.random.normal(size=n)
        for i in range(len(sizes)):
            size = sizes[i]
            color = 0 if MPI.COMM_WORLD.rank < size else 1
            if MPI.COMM_WORLD.size == size:
                comm = MPI.COMM_WORLD
            elif size == 1:
                comm = MPI.COMM_SELF
            else:
                comm = MPI.COMM_WORLD.Split(color)

            pnp = Reduction(comm)
            # PObjective = MPI_Objective_Interface(Objective,
            # nb_domain_grid_pts=n, comm=comm)
            PObjective = MPI_Quadratic(nb_domain_grid_pts=n, pnp=pnp,
                                       factors=factors, startpoint=startpoint)
            x0 = PObjective.startpoint()

            if MPI.COMM_WORLD.Get_rank() == 0:
                print(" Before min n = {}".format(n))

            res[i], t[i] = timer(LBFGS, PObjective.f, x0, jac=PObjective.grad,
                                 maxcor=maxcor, maxiter=100000, gtol=(1e-5),
                                 store_iterates=None, pnp=pnp)
            msg += "size {}:\n".format(size)

            assert res[i].success, "Minimization faild"
            assert pnp.max(
                abs(np.reshape(res[i].x, (-1,)) - np.reshape(PObjective.xmin(),
                                                             (-1,)))) \
                   / pnp.max(
                abs(PObjective.startpoint() - PObjective.xmin())) < 1e-5
            if MPI.COMM_WORLD.Get_rank() == 0:
                print("n = {}, size = {}".format(n, size))

        if toPlot:
            ax.plot(sizes, float(sizes[0] * t[0]) / t, '-o',
                    label="n = {}".format(n))
            ax2.plot(sizes, t, '-o', label="n = {}".format(n))
            fig.savefig("LBFGS_parallel_speedup.png")

    if toPlot:
        ax.legend()
        fig.savefig("LBFGS_parallel_speedup.png")


if __name__ == "__main__":
    show_parallel_speedup()



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import pytest
import numpy as np
import numpy.ma as ma

from NuMPI.Tools import Reduction


@pytest.fixture
def pnp(comm):
    return Reduction(comm)


def test_sum_scalar(pnp):
    res = pnp.sum(np.array(1))
    assert res == pnp.comm.Get_size()


@pytest.mark.parametrize('arr', [
    [1, 2.1, 3],
    np.array((1, 2.1, 3)),
    ma.masked_array([1, 2.1, 3])
])
def test_sum_1D(pnp, arr):
    res = pnp.sum(arr)
    np.testing.assert_allclose(res, pnp.comm.Get_size() * 6.1, atol=1e-12)


@pytest.mark.parametrize('arr', [
    [[1, 2.1, 3],
     [4, 5, 6]],
    np.array(((1, 2.1, 3),
              (4, 5, 6))),
    ma.masked_array([[1, 2.1, 3],
                     [4, 5, 6]])
])
def test_sum_2D(pnp, arr):
    res = pnp.sum(arr)
    np.testing.assert_allclose(res, pnp.comm.Get_size() * 21.1, atol=1e-12)


@pytest.mark.parametrize('arr', [
    np.array(((1, 2.1, 3),
              (4, 5, 6))),
    ma.masked_array([[1, 2.1, 3],
                     [4, 5, 6]])
])
def test_sum_boolean(pnp, arr):
    arr = arr > 3

    # print(arr.dtype)
    res = pnp.sum(arr)
    assert res == pnp.comm.Get_size() * 3


def test_sum_along_axis_decomp(pnp):
    nb_domain_grid_pts = (128, 65)
    np.random.seed(2)
    globaldata = np.random.random(nb_domain_grid_pts)

    nprocs = pnp.comm.Get_size()
    rank = pnp.comm.Get_rank()

    step = nb_domain_grid_pts[0] // nprocs

    if rank == nprocs - 1:
        subdomain_slices = (slice(rank * step, None), slice(None, None))
        # subdomain_locations = [rank * step, 0]
        # nb_subdomain_grid_pts = [nb_domain_grid_pts[0] - rank * step,
        #                         nb_domain_grid_pts[1]]
    else:
        subdomain_slices = (
            slice(rank * step, (rank + 1) * step), slice(None, None))
        # subdomain_locations = [rank * step, 0]
        # nb_subdomain_grid_pts = [step, nb_domain_grid_pts[1]]

    localdata = globaldata[subdomain_slices]

    np.testing.assert_allclose(pnp.sum(localdata, axis=0),
                               np.sum(globaldata, axis=0), rtol=1e-12)
    # Why are they not perfectly equal ?


@pytest.mark.parametrize('arr', [
    np.reshape(np.array((-1, 1, 5, 4,
                         4, 5, 4, 5,
                         7, 0, 1, 0.), dtype=float), (3, 4)),
    ma.masked_array(np.reshape(np.array((-1, 1, 5, 4,
                                         4, 5, 4, 5,
                                         7, 0, 1, 0.), dtype=float), (3, 4)))
])
def test_max_2D(pnp, arr):
    rank = pnp.comm.Get_rank()
    if pnp.comm.Get_size() >= 4:
        if rank == 0:
            local_arr = arr[0:2, 0:2]
        elif rank == 1:
            local_arr = arr[0:2, 2:]
        elif rank == 2:
            local_arr = arr[2:, 0:2]
        elif rank == 3:
            local_arr = arr[2:, 2:]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    elif pnp.comm.Get_size() >= 2:
        if rank == 0:
            local_arr = arr[0:2, :]
        elif rank == 1:
            local_arr = arr[2:, :]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    else:
        local_arr = arr
    assert pnp.max(local_arr) == 7


@pytest.mark.parametrize('arr', [
    np.reshape(np.array((-1, 1, 5, 4,
                         4, 5, 4, 5,
                         7, 0, 1, 0), dtype=int), (3, 4)),
    ma.masked_array(np.reshape(np.array((-1, 1, 5, 4,
                                         4, 5, 4, 5,
                                         7, 0, 1, 0), dtype=int), (3, 4)))
])
def test_max_2D_int(pnp, arr):
    rank = pnp.comm.Get_rank()
    if pnp.comm.Get_size() >= 4:
        if rank == 0:
            local_arr = arr[0:2, 0:2]
        elif rank == 1:
            local_arr = arr[0:2, 2:]
        elif rank == 2:
            local_arr = arr[2:, 0:2]
        elif rank == 3:
            local_arr = arr[2:, 2:]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    elif pnp.comm.Get_size() >= 2:
        if rank == 0:
            local_arr = arr[0:2, :]
        elif rank == 1:
            local_arr = arr[2:, :]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    else:
        local_arr = arr
    assert pnp.max(local_arr) == 7


def test_max_min_empty(pnp):
    """
    Sometimes the input array is empty
    """
    if pnp.comm.Get_size() >= 2:
        if pnp.comm.Get_rank() == 0:
            local_arr = np.array([], dtype=float)

        else:
            local_arr = np.array([1, 0, 4], dtype=float)
        assert pnp.max(local_arr) == 4
        assert pnp.min(local_arr) == 0

        if pnp.comm.Get_rank() == 0:
            local_arr = np.array([1, 0, 4], dtype=float)
        else:

            local_arr = np.array([], dtype=float)
        assert pnp.max(local_arr) == 4
        assert pnp.min(local_arr) == 0

    else:
        local_arr = np.array([], dtype=float)
        # self.assertTrue(np.isnan(pnp.max(local_arr)))
        # self.assertTrue(np.isnan(pnp.min(local_arr)))
        assert pnp.max(local_arr) == np.finfo(local_arr.dtype).min
        assert pnp.min(local_arr) == np.finfo(local_arr.dtype).max


@pytest.mark.parametrize('arr', [
    np.reshape(np.array((-1, 1, 5, 4,
                         4, 5, 4, 5,
                         7, 0, 1, 0), dtype=float), (3, 4)),
    ma.masked_array(np.reshape(np.array((-1, 1, 5, 4,
                                         4, 5, 4, 5,
                                         7, 0, 1, 0), dtype=float), (3, 4)))
])
def test_min(pnp, arr):
    rank = pnp.comm.Get_rank()
    if pnp.comm.Get_size() >= 4:
        if rank == 0:
            local_arr = arr[0:2, 0:2]
        elif rank == 1:
            local_arr = arr[0:2, 2:]
        elif rank == 2:
            local_arr = arr[2:, 0:2]
        elif rank == 3:
            local_arr = arr[2:, 2:]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    elif pnp.comm.Get_size() >= 2:
        if rank == 0:
            local_arr = arr[0:2, :]
        elif rank == 1:
            local_arr = arr[2:, :]
        else:
            local_arr = np.empty(0, dtype=arr.dtype)
    else:
        local_arr = arr
    assert pnp.min(local_arr) == -1


def test_dot_row_vectors(pnp):
    np.random.seed(1)

    n = 10
    fulla = np.random.random(n)
    fullb = np.random.random(n)

    worldsize = pnp.comm.Get_size()
    rank = pnp.comm.Get_rank()
    step = n // worldsize

    if rank == worldsize - 1:
        loc_sl = slice(rank * step, None)
    else:
        loc_sl = slice(rank * step, (rank + 1) * step)

    # because the addition of floating point numbers are not done in the same
    # order due to the parallelization, the truncation errors can be different
    # TODO: maybe we can quantify more seriously what the error should be ?
    np.testing.assert_almost_equal(pnp.dot(fulla[loc_sl], fullb[loc_sl]),
                                   np.dot(fulla, fullb), 7)


def test_dot_matrix_vector(pnp):
    np.random.seed(1)

    m = 13
    n = 11

    worldsize = pnp.comm.Get_size()
    rank = pnp.comm.Get_rank()

    step = n // worldsize

    if rank == worldsize - 1:
        loc_sl = slice(rank * step, None)
    else:
        loc_sl = slice(rank * step, (rank + 1) * step)

    fulla = np.random.random((m, n))
    for fullb in [np.random.random((n, 1)), np.random.random(n)]:
        np.testing.assert_allclose(pnp.dot(fulla[:, loc_sl], fullb[loc_sl]),
                                   np.dot(fulla, fullb))
        np.testing.assert_allclose(
            pnp.dot(fullb[loc_sl].T, fulla[:, loc_sl].T),
            np.dot(fullb.T, fulla.T))
        with pytest.raises(ValueError):
            pnp.dot(fullb[loc_sl], fulla[:, loc_sl])
        with pytest.raises(ValueError):
            pnp.dot(fulla[:, loc_sl].T, fullb[loc_sl])


def test_dot_matrix_matrix(pnp):
    np.random.seed(1)  # important so that every processor sees the same data

    worldsize = pnp.comm.Get_size()
    rank = pnp.comm.Get_rank()

    m = 5
    n = 10
    fulla = np.random.random((m, n))
    fullb = np.random.random((n, m))

    step = n // worldsize

    if rank == worldsize - 1:
        loc_sl = slice(rank * step, None)
    else:
        loc_sl = slice(rank * step, (rank + 1) * step)

    np.testing.assert_allclose(pnp.dot(fulla[:, loc_sl], fullb[loc_sl, :]),
                               np.dot(fulla, fullb))


def test_any_scalar(pnp):
    rank = pnp.comm.Get_rank()

    locval = False

    if rank == 0:
        locval = True

    assert pnp.any(locval)

    locval = False

    assert not pnp.any(locval)


@pytest.mark.parametrize('locval', [
    np.array([False, False, True]),
    ma.masked_array([False, False, True])
])
def test_any_array(pnp, locval):
    rank = pnp.comm.Get_rank()

    if rank == 0:
        locval = np.array([False, True])

    assert pnp.any(locval)

    locval = np.array([False, False, False])

    assert not pnp.any(locval)


def test_all_scalar(pnp):
    rank = pnp.comm.Get_rank()

    locval = True

    if rank == 0:
        locval = False

    assert not pnp.all(locval)

    locval = True

    assert pnp.all(locval)


@pytest.mark.parametrize('locval', [
    np.array([True, True, True]),
    ma.masked_array([True, True, True])
])
def test_all_array(pnp, locval):
    rank = pnp.comm.Get_rank()

    if rank == 0:
        locval = np.array([False, True])

    assert not pnp.all(locval)

    locval = np.array([True, True, True])

    assert pnp.all(locval)



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from .. import MPI


def get_dtype_info(dtype):
    if dtype.kind == 'i':
        return np.iinfo(dtype)
    if dtype.kind == 'f':
        return np.finfo(dtype)
    raise ValueError


class Reduction:
    def __init__(self, comm=None):
        if comm is None:
            self.comm = MPI.COMM_SELF
        else:
            self.comm = comm

    def _op(self, npop, npargs, mpiop, *args, **kwargs):
        """
        Generic reduction operation

        Parameters
        ----------
        npop : func
            Numpy reduction function (e.g. np.sum)
        npargs: tuple
            Arguments passed to the reduction function (e.g. array to be
            reduced)
        mpiop : mpi4py.MPI.op
            MPI reduction operation

        Returns
        -------
        result_arr : np.ndarray
            Result of the reduction operation
        """
        local_result = npop(*npargs, *args, **kwargs)
        result = np.zeros_like(local_result)
        mpitype = MPI._typedict[local_result.dtype.char]
        self.comm.Allreduce([local_result, mpitype], [result, mpitype], op=mpiop)
        return result

    def _op1(self, npop, arr, mpiop, *args, **kwargs):
        """
        Generic reduction operation that takes a single (array) argument

        Parameters
        ----------
        npop : func
            Numpy reduction function (e.g. np.sum)
        arr : array_like
            Numpy array containing the data to be reduced
        mpiop : mpi4py.MPI.op
            MPI reduction operation
        initial : arr.dtype
            Value to use if local array is empty

        Returns
        -------
        result_arr : np.ndarray
            Result of the reduction operation
        """
        if 'initial' in kwargs and isinstance(arr, np.ma.MaskedArray):
            # Max/min on masked array do not support `initial`
            arr = arr.filled(kwargs['initial'])
            del kwargs['initial']
        return self._op(npop, (arr,), mpiop, *args, **kwargs)

    def sum(self, arr, *args, **kwargs):
        """
        Summation

        Parameters
        ----------
        arr : array_like
            Numpy array containing the data to be reduced

        Returns
        -------
        result_arr : np.ndarray
            Sum of all elements of the array over all processors
        """
        return self._op1(np.sum, arr, MPI.SUM, *args, **kwargs)

    def max(self, arr, *args, **kwargs):
        """
        Maximum value

        Parameters
        ----------
        arr : array_like
            Numpy array containing the data to be reduced

        Returns
        -------
        result_arr : np.ndarray
            Maximum of all elements of the array over all processors
        """
        kwargs['initial'] = get_dtype_info(arr.dtype).min
        return self._op1(np.max, arr, MPI.MAX, *args, **kwargs)

    def min(self, arr, *args, **kwargs):
        """
        Minimum value

        Parameters
        ----------
        arr : array_like
            Numpy array containing the data to be reduced

        Returns
        -------
        result_arr : np.ndarray
            Minimum of all elements of the array over all processors
        """
        kwargs['initial'] = get_dtype_info(arr.dtype).max
        return self._op1(np.min, arr, MPI.MIN, *args, **kwargs)

    def mean(self, arr, *args, **kwargs):
        """
        Arithmetic mean

        Parameters
        ----------
        arr : array_like
            Numpy array containing the data to be reduced

        Returns
        -------
        result_arr : np.ndarray
            Arithmetic mean of all elements of the array over all processors
        """
        return self.sum(arr, *args, **kwargs) / self.sum(np.ones_like(arr), *args, **kwargs)

    def dot(self, a, b, *args, **kwargs):
        """
        Scalar product a.b

        Parameters
        ----------
        a : array_like
            Numpy array containing the data of the first array
        a : array_like
            Numpy array containing the data of the second array

        Returns
        -------
        result_arr : np.ndarray
            Scalar product between a and b
        """
        return self._op(np.dot, (a, b), MPI.SUM, *args, **kwargs)

    def any(self, arr, *args, **kwargs):
        """
        Returns true of any value is true

        Parameters
        ----------
        arr : array of bools
            Input data

        Returns
        -------
        result_arr : np.ndarray
            True if any value in `arr` is true
        """
        return self._op1(np.any, arr, MPI.LOR, *args, **kwargs)

    def all(self, arr, *args, **kwargs):
        """
        Returns true of all values are true

        Parameters
        ----------
        arr : array of bools
            Input data

        Returns
        -------
        result_arr : np.ndarray
            True if all values in `arr` are true
        """
        return self._op1(np.all, arr, MPI.LAND, *args, **kwargs)



#
# Copyright 2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from NuMPI.Optimization import LBFGS
from test.Optimization.MPI_minimization_problems import MPI_Quadratic
from NuMPI import MPI
from NuMPI.Tools import Reduction
from memory_profiler import profile
import sys

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
nprocs = comm.Get_size()
pnp = Reduction(comm)
print(sys.argv)
n = int(sys.argv[1])
print(n)

fp = open("LBFGS_memory_profile_n{}_nprocs{}_rank{}.log".format(n, nprocs, rank), "w")
LBFGS = profile(stream=fp)(LBFGS)

Objective = MPI_Quadratic(n, pnp=pnp)
result = LBFGS(Objective.f, Objective.startpoint(),
               jac=Objective.grad,
               options={"gtol": 1e-6, "pnp": Objective.pnp})
fp.close()
assert result.success
print(result.nit)



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import subprocess


class CannotDiscoverVersion(Exception):
    pass


def get_version_from_git():
    """
    Discover version from git repository.
    """
    if not os.path.exists('.git'):
        raise CannotDiscoverVersion('.git subdirectory does not exist.')

    try:
        git_describe = subprocess.run(
            ['git', 'describe', '--tags', '--dirty', '--always'],
            stdout=subprocess.PIPE)
    except FileNotFoundError:
        git_describe = None
    if git_describe is None or git_describe.returncode != 0:
        raise CannotDiscoverVersion('git execution failed.')
    version = git_describe.stdout.decode('latin-1').strip()

    dirty = version.endswith('-dirty')

    # Make version PEP 440 compliant
    if dirty:
        version = version.replace('-dirty', '')
    version = version.strip('v')  # Remove leading 'v' if it exists
    version = version.replace('-', '.dev', 1)
    version = version.replace('-', '+', 1)
    if dirty:
        version += '.dirty'

    return version


_pkg_name = __name__.replace('.DiscoverVersion', '')

# importlib is present in Python >= 3.8
try:
    from importlib.metadata import version

    __version__ = version(_pkg_name)
except ImportError:
    __version__ = None

# git works if we are in the source repository
if __version__ is None:
    try:
        __version__ = get_version_from_git()
    except CannotDiscoverVersion:
        __version__ = None

# Nope. Out of options.

if __version__ is None:
    raise CannotDiscoverVersion('Tried importlib, pkg_resources, PKG-INFO and git')
